
/*
 * save_image.c 
 */
#include <stdio.h>
#include <sys/types.h>

#include <time.h>
#include <sys/time.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include "shm.h"

#include <errno.h>

extern int debug;
extern int verbose;

void init_shm();
void show_sentry_shm();
void save_thru(struct timeval, int, int, char**);

key_t   sentry_shmkey;
int     sentry_shmid;
int     sentry_shm_chunk_radix;
sentry_shmrec   *sentry_shm_chunk;
sentry_shmrec   *sentry_shm_info;

void init_shm()
{
  fprintf(stderr, "init_shm:\n");

  sentry_shmkey = SENTRY_SHMKEY;
  sentry_shm_chunk_radix = SENTRY_SHM_CHUNK_RADIX;

  if ((sentry_shmid = 
       make_new_pool(sentry_shmkey, SENTRY_SHM_BLOCK_SIZE*(sentry_shm_chunk_radix+1))) < 0) {
    perror("init_shm: cannot make shm");
    exit (1);
  }
  if (debug) fprintf(stderr, "ok0: sentry_shmid=%d\n", sentry_shmid);

  if ((sentry_shm_chunk = (sentry_shmrec *)shmat(sentry_shmid, NULL, 0)) == (sentry_shmrec *)-1) {
    perror("init_shm: fail to shmat()");
    exit (1);
  }
  if (debug) fprintf(stderr, "ok1: sentry_shm_chunk=%d\n", sentry_shm_chunk);

  sentry_shm_info = &sentry_shm_chunk[0];
  bzero((char *)sentry_shm_info->info.magick, 16);
  strcpy(sentry_shm_info->info.magick, "SENTRY");
  bzero((char *)sentry_shm_info->info.ver, 0);
  strcpy(sentry_shm_info->info.ver, "0.22");

  sentry_shm_info->info.last_update = -1;
  sentry_shm_info->info.radix = sentry_shm_chunk_radix;
  
  show_pool_stat(sentry_shmid);
  show_sentry_shm();
}

void
show_sentry_shm()
{
  /* no-op */
}

/*
 * save_thru
 */
void
save_thru(struct timeval tstamp, int size, int format, char **buff)
{
  static int h = 0;
  static int c = 0;
  static time_t last = -1;
  time_t now;
  int pos ;

  if (h==0) init_shm();
  h++;

#if 1
  time(&now);
  if (now != last) 
    c = 0;
  else
    c ++;
  last = now;
  if (c != 0) return;
  pos = 1+now%sentry_shm_chunk_radix;
#else
  if (tstamp.tv_sec != last) 
    c = 0;
  else 
    c ++;
  last = tstamp.tv_sec;
  rate = server_rate / 100;
  if  (c > rate) return;

  pos = (tstamp.tv_sec*rate)%sentry_shm_chunk_radix;
#endif

  if (verbose) 
    fprintf(stderr, "save_thru: h %8d, c %3d : now %10d -> pos %4d\n\n",
	    h, c, now, pos);

  sentry_shm_chunk[pos].entry.header.attr.tstamp = -1;

  if (debug)
    fprintf(stderr, "save_thru: size# %d\n", size);

  if (size > SENTRY_SHM_BLOCK_SIZE-16) {
    size = SENTRY_SHM_BLOCK_SIZE - 16;
    return;
  }

  bcopy(*buff, sentry_shm_chunk[pos].entry.body, size);
  /*memcpy(sentry_shm_chunk[pos].entry.body, buff, size);*/
  sentry_shm_chunk[pos].entry.header.attr.length = size;

  sentry_shm_chunk[pos].entry.header.attr.tstamp = now;
  sentry_shm_info->info.last_update = now;
}
