/*
 * Sentry Video System
 *   controls QoS from clients' requests.
 * 
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997, 2000
 *
 * $Id: $
 */

/*
 * main.c - main
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_SYSLOG
#include <syslog.h>
#endif
#include <errno.h>

#include "sentry.h"
#include "sentryd.h"

int debug;
int verb;

#ifdef LOGTIME
time_t logtime;
struct tm *log_tm;
#endif

int data_port = SENTRY_DATA_PORT;
int ctrl_port = SENTRY_CTRL_PORT;

int tcpsock;            /* socket(ctrl/data) for tcp connection */
int udpsock;            /* socket(ctrl only) for udp connection */

#if 0
int unicast_datasock;   /* data socket for udp uniicast */
#endif
int multicast_datasock; /* data socket for udp multicast */

int maxreadfds;
fd_set readmask;

int maxwritefds;
fd_set writemask;

char vserver[80] = "localhost";
int  require_encode_type = ENCODE_NONE;

extern int xl_offset, xr_offset;

int 
bind_socket(int sockfd, int port) 
{
  struct sockaddr_in serv_addr;

  bzero((char *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);

  if (bind(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    fprintf(stderr, "sentryd: cannot bind port %d.\n", port);
    close(sockfd);
    return -1;
  }
  return 0;
}

int 
build_listen_port(int port, int type)
{
  int sockfd;
  int flag=1;

  while (1) {
    if ((sockfd = socket(AF_INET, type, 0)) < 0) {
      fprintf(stderr, "sentryd: cannot make new TCP socket.\n");
      return -1;
    }
    
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag)) < 0) {
      fprintf(stderr, "sentryd: cannot set sockopt REUSEADDR.\n");
      /* return -1; */
    }

    if (bind_socket(sockfd, port) < 0) {
      fprintf(stderr, "sentryd: please wait several seconds.\n");
      close(sockfd);
      sleep(3);
      continue;
    }
    break;
  }

  listen(sockfd, -1);
  if (verb)
    fprintf(stderr, "sentryd: new socket# %d, type# %d, port#, %d.\n", sockfd, type, port);

  return sockfd;
}

int
build_udp_port(int port) 
{
  int sockfd;
  int flag=1;
  int flaglen = sizeof(flag);

  if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "sentryd: cannot make new UDP socket.\n");
    return -1;
  }

  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
                 (char *)&flag, sizeof(flag)) < 0) {
    fprintf(stderr, "sentryd: cannot setsockopt REUSEADDR\n");
    return -1;
  }


  /*
   * set SNDBUF size:
   *   challenge 16M, 8M, ...., 16K, 8K.
   */
  flag = M(16);
  while(flag >= K(8)) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                 (char *)&flag, flaglen) < 0) {
#if 0
      if (debug) 
	fprintf(stderr, "sentryd: fail set SNDBUF to %9d bytes.\n", flag);
#endif
      flag = flag / 2 ;
    }
    else {
      fprintf(stderr, "sentryd: success set SNDBUF size [%9d bytes].\n", flag);
      break;
    }
  }

  if (verb)
    if (getsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                   (char *)&flag, &flaglen) < 0)
      fprintf(stderr, "sentryd: cannot get SNDBUF size.\n");
    else
      fprintf(stderr, "sentryd: SNDBUF size = %d\n", flag);
  

  if (bind_socket(sockfd, port) < 0) {
    fprintf(stderr, "sentryd: please wait several seconds.\n");
    close(sockfd);
    return -1;
  }

  if (verb)
    fprintf(stderr, "sentryd: new udp socket# %d, port# %d.\n", sockfd, port);

  return sockfd;
}


/*
 * connect_host(): create a connect-mode socket.
 *   addr: network byte order address
 *   port: network byte order port# 
 *   type: SOCK_DGRAM/SOCK_STREAM, now only supported SOCK_DGRAM.
 *   return value: socket discripter
 */
int
connect_host(struct in_addr addr, int port, int type)
{
  int sockfd;
  struct sockaddr_in peer;

  switch(type) {
  case SOCK_DGRAM:
    if ((sockfd = build_udp_port(0)) < 0) return -1;
    break;
  case SOCK_STREAM:
  default:
    return -1;
  }

  peer.sin_family = AF_INET;
  peer.sin_addr.s_addr = addr.s_addr;
  peer.sin_port = port;
  if (connect(sockfd, (struct sockaddr *)&peer, sizeof(peer)) < 0) {

    fprintf(stdout, "connect_host: fail to connect to %s:%d\n",
	    inet_ntoa(peer.sin_addr.s_addr),
	    ntohs(peer.sin_port));
    close(sockfd);
    return -1;
  }

#if 0
  if (0) {
    char buf[128];
    int nw = 0;
    sprintf(buf, "connect_host: connected to %s:%d\n",
	    inet_ntoa(peer.sin_addr), ntohs(port));
    nw = write(sockfd, buf, strlen(buf));
    fprintf(stderr, "nw = %d\n", nw);

    fprintf(stderr, buf);
  }
#endif

  return sockfd;
}



void
reaper()
{
  union wait status;
  while(wait3((int *)&status, WNOHANG, (struct rusage *)0) >= 0)
    ;
}

void
finish_sentryd()
{
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif

  if (tcpsock >= 0) {
    shutdown(tcpsock, 2);
    close(tcpsock);
    tcpsock = -1;
  }
  if (udpsock >=0) {
    shutdown(udpsock, 2);
    close(udpsock);
    udpsock = -1;
  }
  if (multicast_datasock >= 0) {
    shutdown(multicast_datasock, 2);
    close(multicast_datasock);
    multicast_datasock = -1;
  }

  if (verb) {
#ifdef LOGTIME
    logtime = time(NULL);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
	    log_tm->tm_year + 1900, log_tm->tm_mon+1, log_tm->tm_mday,
	    log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "finish sentryd\nBye.\n");
  }

#ifdef USE_SYSLOG
  syslog(LOG_NOTICE, "sentryd is stopped.\n");
  closelog();
#endif

  exit (0);
}

void
mainloop()
{
  fd_set readfds;
  int nfds;
  char sysmesg[256];
  int newsockfd;
  struct sockaddr_in from;
  int fromlen;
  int nr;
  char buf[8192];
  extern void init_req_val(), inti_client();

  (void)init_req_val();
  (void)init_client();

  fromlen = sizeof(from);

  FD_ZERO(&readmask);
  FD_SET(tcpsock, &readmask); 
  maxreadfds = tcpsock;
  FD_SET(udpsock, &readmask);
  if (maxreadfds < udpsock) maxreadfds = udpsock;

  FD_ZERO(&writemask);
  maxwritefds = 0;
  /*
  FD_SET(multicast_datasock, &writemask);
  maxwritefds = multicast_datasock;
  */

  while (1) {
    readfds = readmask;

    nfds = select(maxreadfds+1, &readfds, (fd_set *)NULL, (fd_set *)NULL, (struct timeval *)NULL);
    if (nfds <= 0) {
#ifdef USE_SYSLOG
      sprintf(sysmesg, "sentryd: select: %s", sys_errlist[errno]);
      syslog(LOG_ERR, sysmesg);
#endif
      continue;
    }

    if (FD_ISSET(tcpsock, &readfds)) { /* connection request arrived at tcp port */
      FD_CLR(tcpsock, &readfds);
      newsockfd = accept(tcpsock, (struct sockaddr *)&from, &fromlen);
      if (newsockfd < 0)
	goto udp_check;

      if (verb) 
	fprintf(stderr, "TCP connection is arrived.\n");

      if (debug) {
	fprintf(stderr, "from->sin_addr = %s\n", inet_ntoa(from.sin_addr));
	fprintf(stderr, "from->sin_port = %d\n", 
		ntohs(from.sin_port));
      }

      /* get_tcp_request(newsockfd) */

      if (add_client(0xf2, &from, newsockfd) < 0) {
	fprintf (stderr, "close connection port# %d.\n", newsockfd);
	shutdown(newsockfd, 2);
	close(newsockfd);
	continue;
      }

      FD_SET(newsockfd, &readmask);
      if (maxreadfds < newsockfd) maxreadfds = newsockfd;

      FD_SET(newsockfd, &writemask);
      if (maxwritefds < newsockfd) maxwritefds = newsockfd;
      
      fprintf(stdout, "sentryd: tcp sock# %d is connected from %s:%d.\n", 
	      newsockfd,
	      inet_ntoa((struct in_addr *)ntohl(from.sin_addr.s_addr)),
	      ntohs(from.sin_port));
      continue;
    }

  udp_check:
    if (FD_ISSET(udpsock, &readfds)) { /* request arrived at udp port */
      if (verb)
	fprintf(stderr, "UDP connection/request is arrived.\n");

      get_udp_request(&from, &fromlen);

      if (debug) {
	u_long addr = ntohl(from.sin_addr.s_addr);
	fprintf(stderr, "addr# %x -> %x, %s\n", 
		ntohl(from.sin_addr.s_addr), addr,
		inet_ntoa((struct in_addr *)ntohl(addr)));
      }

      printf("sentryd: QoS request are arrived to udp port# %d from %s:%d\n", 
	     udpsock, inet_ntoa(from.sin_addr), ntohs(from.sin_port));

    }else { /* QoS request arrived at TCP port */
      /*
       * QoS requests are arrived from TCP port.
       *   not implemented yet... :)
       */
      if (verb) 
	fprintf(stdout, "TCP request is arrived.\n");

      fprintf(stderr, "this routine has not implemented yet.\n");
      continue;
      /*
      get_tcp_request();
      */
    }

  } /* end-of-while */
  /* NOT REACHED */
}


int
main(int argc, char **argv)
{
  int c;
  extern char *optarg;
  extern void usage(), finish_sentryd();
#ifdef LOGTIME
  time_t logtime;
  struct tm *log_tm;
#endif /* LOGTIME */
#ifdef DEBUG
  debug = 1;
#endif /* DEBUG */
#ifdef VERBOSE
  verb = 1;
#endif /* VERBOSE */

#ifdef USE_SYSLOG
  openlog("sentryd", LOG_PID, LOG_DAEMON);
  syslog(LOG_NOTICE, "sentryd starts.");
#endif

  if (verb) {
#ifdef LOGTIME
    logtime = time(null);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
	    log_tm->tm_year+1900, log_tm->tm_mon+1, log_tm->tm_mday,
	    log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stdout, "sentryd starts.\n");
  }

  signal(SIGQUIT, finish_sentryd);
  signal(SIGTERM, finish_sentryd);
  signal(SIGINT, finish_sentryd);

  while ((c = getopt(argc, argv, "hDvjp:l:r:")) != -1) {
    switch(c) {
    case 'h':
      usage();
      exit (1); /* not reached */
    case 'D':
      debug ++;
      break;
    case 'v':
      verb = 1;
      break;
    case 'j':
      require_encode_type = ENCODE_JPEG;
      break;
    case 'p':
      data_port = atoi(optarg);
      ctrl_port = data_port + 1;
      break;
#ifndef IRIX  /* Linux */
    case 'l':
      xl_offset = atoi(optarg);
      if(xl_offset < 0) xl_offset = 0;
      break;
    case 'r':
      xr_offset = atoi(optarg);
      if(xr_offset < 0) xr_offset = 0;
      break;
#else  /* IRIX */
    case 'l':
    case 'r':
      usage();
      exit(1);
#endif
    }
  }
  
  /*
   * video setup
   */
  if (check_video(vserver)) {
    fprintf(stderr, "sentryd: sorry, video(%s) is not avaliable.\n", "localhost");
    exit (2);
  }

  tcpsock = build_listen_port(ctrl_port, SOCK_STREAM);
  udpsock = build_listen_port(ctrl_port, SOCK_DGRAM);

  multicast_datasock = build_udp_port(0);

  if (tcpsock < 0 || udpsock < 0)
    exit (1);
  
  (void)mainloop();
  exit (0);
}

void
usage() {
  printf("usage: sentryd [options]\n\n");

  printf("option:\n");
  printf("       -h       help\n");
  printf("       -j       enable jpeg encode(default=disable)\n");
  printf("       -p port  data port(default=%d)\n", SENTRY_DATA_PORT);
  printf("       -D       debug mode\n");
  printf("       -v       verbose mode\n");
  printf("       -l       left side offset(default=0)\n");
  printf("       -r       right side offset(default=0)\n");
  printf("       At IRIX, you can't use l and r options\n");
  printf("\n");

  exit (1);
}



  
  
